"""
This module contains cli commands for pyaki.

To use the CLI you can use the `pyaki-cli` command.
For more information on how to use the CLI, run `pyaki-cli --help`.

```bash
Usage: pyaki-cli [OPTIONS] PATH

CLI tool to process AKI stages from time series data.

Arguments:
    *    path      TEXT  [default: None] [required]

Options:
    --urineoutput-file         TEXT  [default: urineoutput.csv]
    --rrt-file                 TEXT  [default: rrt.csv]
    --demographics-file        TEXT  [default: demographics.csv]
    --help                           Show this message and exit.
```
"""



"""
This module contains the preprocessors used in the pyaki package.
"""

from abc import ABC
from typing import Optional

import pandas as pd
from pandas.api.types import is_datetime64_any_dtype

from pyaki.utils import Dataset, DatasetType, dataset_as_df, df_to_dataset


class Preprocessor(ABC):
    """
    Abstract base class for preprocessors.

    Parameters
    ----------
    stay_identifier : str, default: "stay_id"
        The column name that identifies stays or admissions in the dataset.
    time_identifier : str, default: "charttime"
        The column name that identifies the timestamp or time variable in the dataset.
    """

    def __init__(self, stay_identifier: str = "stay_id", time_identifier: str = "charttime") -> None:
        super().__init__()

        self._stay_identifier: str = stay_identifier
        self._time_identifier: str = time_identifier

    def process(self, datasets: list[Dataset]) -> list[Dataset]:
        """
        Process the given list of datasets and return the processed datasets.

        Parameters
        ----------
        datasets : list[Dataset]
            The list of datasets to be processed.

        Returns
        -------
        list[Dataset]
            The processed datasets.
        """
        raise NotImplementedError()


class TimeIndexCreator(Preprocessor):
    """
    Preprocessor for creating a time index in the datasets.

    Attributes
    ----------
    DATASETS : list[DatasetType]
        The list of dataset types that require a time index.
    """

    DATASETS: list[DatasetType] = [
        DatasetType.CREATININE,
        DatasetType.URINEOUTPUT,
        DatasetType.RRT,
    ]

    def process(self, datasets: list[Dataset]) -> list[Dataset]:
        """
        Process the datasets by creating a time index if the dataset type requires it.

        Parameters
        ----------
        datasets : list[Dataset]
            The list of datasets to be processed.

        Returns
        -------
        list[Dataset]
            The processed datasets.
        """
        _datasets = []
        for dtype, df in datasets:
            if dtype not in self.DATASETS or self._time_identifier not in df.columns:
                _datasets.append(Dataset(dtype, df))
                continue

            if not is_datetime64_any_dtype(df[self._time_identifier]):
                df[self._time_identifier] = pd.to_datetime(df[self._time_identifier])

            _datasets.append(Dataset(dtype, df.set_index(self._time_identifier)))

        return _datasets


class UrineOutputPreProcessor(Preprocessor):
    """
    Preprocessor for processing the urine output dataset.

    Parameters
    ----------
    stay_identifier : str, default: "stay_id"
        The column name that identifies stays or admissions in the dataset.
    time_identifier : str, default: "charttime"
        The column name that identifies the timestamp or time variable in the dataset.
    urineoutput_column : str, default: "urineoutput"
        The column name that represents the urine output values in the dataset.
    interpolate : bool, default: True
        Flag indicating whether to perform interpolation on missing values.
    threshold : int, default: 6
        The threshold value for limiting the interpolation range.
    """

    def __init__(
        self,
        stay_identifier: str = "stay_id",
        time_identifier: str = "charttime",
        urineoutput_column: str = "urineoutput",
        interpolate: bool = True,
        threshold: int = 6,
    ) -> None:
        super().__init__(stay_identifier, time_identifier)
        self._interpolate: bool = interpolate
        self._threshold: int = threshold
        self._urineoutput_column: str = urineoutput_column

    @dataset_as_df(df=DatasetType.URINEOUTPUT)
    @df_to_dataset(DatasetType.URINEOUTPUT)
    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Process the urine output dataset by resampling, interpolating missing values, and applying threshold-based adjustments.

        Parameters
        ----------
        df : pd.DataFrame
            The input urine output dataset as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            The processed urine output dataset as a pandas DataFrame.
        """

        df = df.groupby(self._stay_identifier).resample("1h").sum()  # type: ignore
        df[df[self._urineoutput_column] == 0] = None

        if not self._interpolate:
            return df

        mask = df[self._urineoutput_column].isnull()
        df[self._urineoutput_column] /= (
            (mask.cumsum() - mask.cumsum().where(~mask).ffill().fillna(0))
            .shift(1)
            .clip(upper=self._threshold)
            .add(1)
            .fillna(1)
        )
        return df.bfill(limit=self._threshold)


class CreatininePreProcessor(Preprocessor):
    """
    Preprocessor for processing the creatinine dataset.

    Parameters
    ----------
    stay_identifier : str, default: "stay_id"
        The column name that identifies stays or admissions in the dataset.
    time_identifier : str, default: "charttime"
        The column name that identifies the timestamp or time variable in the dataset.
    creatinine_column : str, default: "creat"
        The column name that represents the creatinine values in the dataset.
    ffill : bool, default: True
        Flag indicating whether to perform forward filling on missing values.
    threshold : int, default: 72
        The threshold value for limiting the forward filling range.
    """

    def __init__(
        self,
        stay_identifier: str = "stay_id",
        time_identifier: str = "charttime",
        creatinine_column: str = "creat",
        ffill: bool = True,
        threshold: int = 72,
    ) -> None:
        super().__init__(stay_identifier, time_identifier)

        self._ffill: bool = ffill
        self._threshold: Optional[int] = threshold
        self._creatinine_column: str = creatinine_column

    @dataset_as_df(df=DatasetType.CREATININE)
    @df_to_dataset(DatasetType.CREATININE)
    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Process the creatinine dataset by resampling and performing forward filling on missing values.

        Parameters
        ----------
        df : pd.DataFrame
            The input creatinine dataset as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            The processed creatinine dataset as a pandas DataFrame.
        """
        df = df.groupby(self._stay_identifier).resample("1h").mean()  # type: ignore
        if not self._ffill:
            return df

        df[df[self._creatinine_column] == 0] = None
        return df.ffill(limit=self._threshold)


class DemographicsPreProcessor(Preprocessor):
    """Preprocessor for processing the demographics dataset."""

    @dataset_as_df(df=DatasetType.DEMOGRAPHICS)
    @df_to_dataset(DatasetType.DEMOGRAPHICS)
    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Process the demographics dataset by aggregating the data based on stay identifiers.

        Parameters
        ----------
        df : pd.DataFrame
            The input demographics dataset as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            The processed demographics dataset as a pandas DataFrame.
        """
        return df.groupby(self._stay_identifier).last()


class RRTPreProcessor(Preprocessor):
    """Preprocessor for processing the RRT dataset."""

    @dataset_as_df(df=DatasetType.RRT)
    @df_to_dataset(DatasetType.RRT)
    def process(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Process the RRT dataset by upsampling the data and forward filling the last value. We expect the dataframe to contain a 1 for RRT in progress, and 0 for RRT not in progress.

        Parameters
        ----------
        df : pd.DataFrame
            The input RRT dataset as a pandas DataFrame.

        Returns
        -------
        pd.DataFrame
            The processed RRT dataset as a pandas DataFrame.
        """
        df = df.groupby(self._stay_identifier).resample("1h").last()  # type: ignore
        return df.ffill()



"""
This module contains the analysis class for processing AKI stages from time series data.
"""

import logging
from typing import Optional

import pandas as pd

from pyaki.preprocessors import (
    CreatininePreProcessor,
    DemographicsPreProcessor,
    Preprocessor,
    RRTPreProcessor,
    TimeIndexCreator,
    UrineOutputPreProcessor,
)
from pyaki.probes import (
    AbsoluteCreatinineProbe,
    Probe,
    RelativeCreatinineProbe,
    RRTProbe,
    UrineOutputProbe,
)
from pyaki.utils import Dataset

logger = logging.getLogger(__name__)


class Analyser:
    """
    Class for data analysis using probes and preprocessors.

    This class provides functionality for analyzing data using a collection of probes and preprocessors.
    It processes the input data through the specified preprocessors and applies the probes to perform
    the analysis. The analysis results are returned as a DataFrame.

    Parameters
    ----------
    data : list[Dataset]
        A list of Dataset objects containing the input data.
    probes : list[Probe], optional
        A list of Probe objects representing the analysis probes to apply. If not provided, default
        probes including UrineOutputProbe, AbsoluteCreatinineProbe, and RelativeCreatinineProbe will be used.
    preprocessors : list[Preprocessor], optional
        A list of Preprocessor objects representing the preprocessors to apply on the input data. If not
        provided, default preprocessors including UrineOutputPreProcessor, CreatininePreProcessor, and
        DemographicsPreProcessor will be used.
    stay_identifier : str, default: "stay_id"
        The column name in the input data representing the stay identifier.
    time_identifier : str, default: "charttime"
        The column name in the input data representing the time identifier.

    Examples
    --------
    Instantiate the Analyser class with custom data, probes, and preprocessors
    ```pycon
    >>> analyser = Analyser(data=my_datasets, probes=[MyProbe()], preprocessors=[MyPreprocessor()])
    ```

    Process stays and obtain the analysis results
    ```pycon
    >>> result_df = analyser.process_stays()
    ```
    """

    def __init__(
        self,
        data: list[Dataset],
        probes: Optional[list[Probe]] = None,
        preprocessors: Optional[list[Preprocessor]] = None,
        stay_identifier: str = "stay_id",
        time_identifier: str = "charttime",
    ) -> None:
        if probes is None:  # apply default probes if not provided
            probes = [
                UrineOutputProbe(),
                AbsoluteCreatinineProbe(),
                RelativeCreatinineProbe(),
                RRTProbe(),
            ]
        if preprocessors is None:  # apply default preprocessors if not provided
            preprocessors = [
                TimeIndexCreator(stay_identifier=stay_identifier, time_identifier=time_identifier),
                UrineOutputPreProcessor(stay_identifier=stay_identifier, time_identifier=time_identifier),
                CreatininePreProcessor(stay_identifier=stay_identifier, time_identifier=time_identifier),
                DemographicsPreProcessor(stay_identifier=stay_identifier),
                RRTPreProcessor(stay_identifier=stay_identifier, time_identifier=time_identifier),
            ]

        # validate datasets
        self.validate_data(data)

        # apply preprocessors to the input data
        logger.info("Start preprocessing")
        for preprocessor in preprocessors:
            data = preprocessor.process(data)

        logger.info("Finish preprocessing")

        self._data: list[Dataset] = data
        self._probes: list[Probe] = probes
        self._stay_identifier: str = stay_identifier

    def validate_data(self, datasets: list[Dataset]) -> None:
        """
        validate the input data for negative values.

        Parameters
        ----------
        datasets : list[Dataset]
            A list of Dataset objects containing the input data.

        Raises
        ------
        ValueError
            If any of the datasets contain negative values.
        """
        for dtype, df in datasets:
            try:
                if (df < 0).values.any():
                    raise ValueError(f"Dataset of Type {dtype} contains negative data")
            except TypeError:
                continue

    def process_stays(self) -> pd.DataFrame:
        """
        Process all stays in the input data.

        This method processes all stays in the input data by applying the configured probes.
        The analysis results for all stays are concatenated and returned as a single DataFrame.

        Returns
        -------
        pd.DataFrame
            The analysis results for all stays.
        """
        logger.info("Start probing")

        (_, df), *datasets = self._data
        stay_ids: pd.Index = df.index.get_level_values(self._stay_identifier).unique()
        for _, df in datasets:
            stay_ids.join(df.index.get_level_values(self._stay_identifier).unique())

        data: pd.DataFrame = self.process_stay(stay_ids.values[0])
        for stay_id in stay_ids.values[1:]:
            data = pd.concat([data, self.process_stay(stay_id)])

        logger.info("Finish probing")
        return data

    def process_stay(self, stay_id: str) -> pd.DataFrame:
        """
        Process a specific stay in the input data by patient identificator.

        This method processes a specific stay in the input data by applying the configured probes and preprocessors.
        The analysis results for the stay are returned as a DataFrame.

        Parameters
        ----------
        stay_id : str
            The identifier of the stay to process.

        Returns
        -------
        pd.DataFrame
            The analysis results for the specific stay.
        """
        logger.debug("Processing stay with id: %s", stay_id)

        datasets: list[Dataset] = [
            Dataset(dtype, data.loc[stay_id])  # type: ignore
            for dtype, data in self._data
            if stay_id in data.index
        ]

        for probe in self._probes:
            datasets = probe.probe(datasets)

        (_, df), *datasets = datasets
        for _, _df in datasets:
            if isinstance(_df, pd.Series):
                _df = pd.DataFrame([_df], index=df.index)
            columns = set(_df.columns) - set(df.columns)
            df = df.merge(_df[[*columns]], how="outer", left_index=True, right_index=True)

        df["stage"] = df.filter(like="stage").max(axis=1)
        return df.set_index(
            pd.MultiIndex.from_arrays(
                [[stay_id] * len(df), df.index.values],
                names=(self._stay_identifier, df.index.name),
            )
        )



#! /usr/bin/env python3
"""pyaki CLI tool to process AKI stages from time series data."""

from pathlib import Path

import pandas as pd
import typer

from pyaki.kdigo import Analyser
from pyaki.utils import Dataset, DatasetType


def main(
    path: str,
    urineoutput_file: str = "urineoutput.csv",
    creatinine_file: str = "creatinine.csv",
    rrt_file: str = "rrt.csv",
    demographics_file: str = "demographics.csv",
) -> None:
    """
    CLI tool to process AKI stages from time series data.

    CLI tool to process AKI stages from time series data. The tool expects
    the following files to be present in the given path: urineoutput.csv, creatinine.csv, rrt.csv, demographics.csv.

    Parameters
    ----------
    path : str
        Path to the folder containing the data files.
    urineoutput_file : str, default: "urineoutput.csv"
        Name of the file containing urine output data.
    creatinine_file : str, default: "creatinine.csv"
        name of the file containing creatinine data.
    rrt_file : str, default: "rrt.csv"
        Name of the file containing rrt data.
    demographics_file : str, default: "demographics.csv"
        Name of the file containing demographic data of the patient like the patients weight.
    """
    root_dir = Path(path)
    datasets = []

    if (ou_file := root_dir / urineoutput_file).is_file():
        datasets.append(Dataset(DatasetType.URINEOUTPUT, pd.read_csv(ou_file)))

    if (scr_file := root_dir / creatinine_file).is_file():
        datasets.append(Dataset(DatasetType.CREATININE, pd.read_csv(scr_file)))

    if (_rrt_file := root_dir / rrt_file).is_file():
        datasets.append(Dataset(DatasetType.RRT, pd.read_csv(_rrt_file)))

    if (demo_file := root_dir / demographics_file).is_file():
        datasets.append(Dataset(DatasetType.DEMOGRAPHICS, pd.read_csv(demo_file)))

    ana: Analyser = Analyser(datasets)
    ana.process_stays().to_csv(root_dir / "aki.csv")


def run() -> None:
    """Run the CLI tool"""
    typer.run(main)


if __name__ == "__main__":
    run()



"""
pyaki: A Python package for classification of acute kidney injury.

This package provides a Python API and CLI tool for analysing time series to classification of acute kidney injury.

Modules:
- bin: Command line interface for the pyaki package.
- kdigo: Implementation of the KDIGO criteria for classification of acute kidney injury.
- preprocessing: Preprocessing of time series data.
- probes: Implementation of the probes for classification of acute kidney injury.
- utils: Utility functions for the pyaki package.

Usage:
```python
import pandas as pd

from pyaki.kdigo import Analyser
from pyaki.probes import Dataset, DatasetType

datasets: list[Dataset] =[
    Dataset(DatasetType.URINEOUTPUT, pd.DataFrame()),
    Dataset(DatasetType.CREATININE, pd.DataFrame()),
    Dataset(DatasetType.DEMOGRAPHICS, pd.DataFrame()),
    Dataset(DatasetType.RRT, pd.DataFrame()),
]

analyser = Analyser(datasets)
reult_df = analyser.process_stays()
```

Author:
- Christian Porschen
- Jan Ernsting
- Paul Brauckmann

License:
```
MIT License

Copyright (c) 2023 AIDH MS

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
```
"""



"""
This module contains the utility functions and classes used in the pyaki package.
"""

import logging
from enum import StrEnum, auto
from functools import wraps
from typing import Any, Callable, NamedTuple, cast

import numpy as np
import pandas as pd

logger = logging.getLogger(__name__)


class DatasetType(StrEnum):
    """
    Enumeration class representing different types of datasets.

    Attributes
    ----------
    URINEOUTPUT : str
        The urine output dataset.
    CREATININE : str
        The creatinine dataset.
    DEMOGRAPHICS : str
        The demographics dataset.
    RRT : str
        The renal replacement therapy dataset.
    """

    URINEOUTPUT = auto()
    CREATININE = auto()
    DEMOGRAPHICS = auto()
    RRT = auto()


class Dataset(NamedTuple):
    """
    Named tuple representing a dataset.

    This class represents a dataset consisting of a dataset type and a DataFrame.
    It is defined as a named tuple, which is an immutable data structure with
    named fields. The `dataset_type` field is of type `DatasetType` and represents
    the type of the dataset. The `df` field is of type `pd.DataFrame` and represents
    the actual data stored in a pandas DataFrame object.

    Attributes
    ----------
    dataset_type : DatasetType
        The type of the dataset.
    df : pd.DataFrame
        The DataFrame object containing the dataset.

    Examples
    --------
    ```pycon
    >>> dataset = Dataset(dataset_type=DatasetType.URINEOUTPUT, df=my_dataframe)
    ```
    """

    dataset_type: DatasetType
    df: pd.DataFrame


def dataset_as_df(**mapping: DatasetType) -> Callable:
    """
    Decorator factory for methods that process datasets with dataframes.

    This decorator is intended to be used with methods in a class that handle datasets
    consisting of dataframes. It allows you to specify a mapping of dataset types to
    corresponding dataframe names. The decorator then replaces the dataframes of the
    specified types with the results of the decorated method.

    Parameters
    ----------
    **mapping : dict[str, DatasetType]
        A mapping of dataset type names to their corresponding `DatasetType`.
        Dataset types not found in this mapping will be ignored.

    Returns
    -------
    decorator : Callable
        A decorator that can be applied to methods in a class. The decorated
        method is expected to accept a list of `Dataset` objects and optional
        additional arguments and keyword arguments.

    Examples
    --------
    Suppose you have a method `process_data` that takes a list of `Dataset` objects
    and a `mapping` as specified in the decorator:

    ```pycon
    >>> @dataset_as_df(data=DatasetType.DATA, labels=DatasetType.LABELS)
    ... def process_data(self, data: pd.DataFrame, labels: pd.DataFrame):
    ...     # Your data processing logic here
    ...     return processed_data, labels
    ```

    When you call `process_data` with a list of `Dataset` objects containing data
    and labels, the decorator will automatically replace the dataframes based on
    the mapping and pass them to the method:

    ```pycon
    >>> processed_datasets = my_instance.process_data(datasets)
    ```
    """
    # swap keys and values in the mapping
    in_mapping: dict[DatasetType, str] = {}
    for k, v in mapping.items():
        in_mapping[cast(DatasetType, v)] = k

    # in_mapping: Dict[DatasetType, str] = {v: k for k, v in mapping.items()}

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(self: Any, datasets: list[Dataset], *args: Any, **kwargs: Any) -> list[Dataset]:
            # map the dataset types to corresponding DataFrames
            _mapping: dict[str, pd.DataFrame] = {
                in_mapping[dtype]: df for dtype, df in datasets if dtype in in_mapping.keys()
            }
            # check if all datasets are mapped, otherwise return the original datasets
            if len(in_mapping) != len(_mapping):
                logger.warning(
                    "Skip %s because one or more datasets are missing to probe",
                    self.__class__.__name__,
                )
                return datasets

            # call the wrapped function with the converted DataFrames
            _dtype, _df = func(self, *args, **_mapping, **kwargs)

            # return the updated datasets
            return [Dataset(dtype, _df if dtype == _dtype else df) for dtype, df in datasets]

        return wrapper

    return decorator


def df_to_dataset(dtype: DatasetType) -> Callable:
    """
    Decorator that converts a DataFrame into a dataset with the specified type.

    This decorator is intended to be used with a method that returns a DataFrame.
    It wraps the method and converts the returned DataFrame into a dataset object
    with the specified type. The converted dataset is then returned.

    Parameters
    ----------
    dtype : DatasetType
        The DatasetType enum value representing the type of the dataset.

    Returns
    -------
    decorator : Callable
        A decorated function that takes the original arguments, performs the wrapped
        function, converts the returned DataFrame into a Dataset object with the
        specified type, and returns the converted dataset.

    Examples
    --------
    ```pycon
    >>> @df_to_dataset(DatasetType.URINEOUTPUT)
    ... def process_dataframe(self, *args: list, **kwargs: dict) -> pd.DataFrame:
    ...     # Process the DataFrame
    ...     ...
    ```
    """

    def decorator(func: Callable) -> Callable:
        @wraps(func)
        def wrapper(self: Any, *args: Any, **kwargs: Any) -> Dataset:
            return Dataset(dtype, func(self, *args, **kwargs))

        return wrapper

    return decorator


def approx_gte(x: pd.Series, y: pd.Series | float) -> bool | np.ndarray:
    """
    Check if x is greater than or approximately equal to y.

    Parameters
    ----------
    x : pd.Series
        The series to compare.
    y : pd.Series or float
        The series or float to compare with.
    """
    return np.logical_or(np.asarray(x >= y), np.isclose(x, y))



"""
This module contains the Probe abstract base class and its subclasses,
used for probing for the different KDIGO criteria.
"""

from abc import ABC, ABCMeta
from enum import StrEnum, auto
from typing import Any

import numpy as np
import pandas as pd
from pandas import PeriodIndex

from pyaki.utils import Dataset, DatasetType, approx_gte, dataset_as_df, df_to_dataset


class Probe(ABC):
    """
    Abstract base class representing a data analysis probe.

    This class serves as an abstract base class (ABC) for data analysis probes.
    It declares the abstract method `probe()` that must be implemented by its subclasses.
    The `RESNAME` attribute can be overridden by subclasses to specify the name of the
    result column generated by the probe.

    Attributes
    ----------
    RESNAME : str
        The name of the result column generated by the probe.

    Methods
    -------
    probe()
        Abstract method to be implemented by subclasses. It performs data analysis on the
        provided datasets and returns a DataFrame with the analysis results.

    Example
    -------
    ```pycon
    >>> class MyProbe(Probe):
    ...     RESNAME = "my_result"
    ...
    ...     def probe(self, datasets: list[Dataset], **kwargs) -> pd.DataFrame:
    ...         # Implementation of the probe's analysis...
    ...
    ... my_probe = MyProbe()
    ... result_df = my_probe.probe(datasets=my_datasets, additional_arg=value)
    ```
    """

    RESNAME: str = ""  # name of the column that will be added to the dataframe

    def probe(self, datasets: list[Dataset], **kwargs: Any) -> list[Dataset]:
        """
        Abstract method to be implemented by subclasses.

        This method performs data analysis on the provided datasets and returns a DataFrame
        with the analysis results. Subclasses must override this method.

        Parameters
        ----------
        datasets : list[Dataset]
            A list of Dataset objects containing the input data.
        **kwargs
            Additional keyword arguments for the analysis.

        Returns
        -------
        pd.DataFrame
            The DataFrame containing the analysis results.
        """
        raise NotImplementedError()


class UrineOutputMethod(StrEnum):
    """
    Enumeration class representing different methods for urine output calculations

    Attributes
    ----------
    STRICT : str
        Strict method for urine output calculations. Using this method, the urine output stage is calculated based on the maximum urine output in the past 6, 12, and 24 hours.
    MEAN : str
        Mean method for urine output calculations.
    """

    STRICT = auto()
    MEAN = auto()


class UrineOutputProbe(Probe):
    """
    Subclass of Probe representing a probe calculating KDIGO stages according to urine output.

    This class specializes the abstract base class `Probe` to perform calculations of KDIGO stages based on urine output. Common KDIGO criteria apply.
    It overrides the `RESNAME` attribute to set the name of the result column.
    The `probe()` method performs urine output analysis on the provided DataFrame and returns a modified DataFrame
    with a column containing the appropriate KDIGO stage, according to urine output, added.

    Attributes
    ----------
    RESNAME : str
        The name of the result column representing urine output stage.

    Parameters
    ----------
    column : str, default: "urineoutput"
        The name of the column representing urine output in the DataFrame.
    patient_weight_column : str, default: "weight"
        The name of the column representing the patient's weight in the patient DataFrame.
    anuria_limit : float, default: 0.1
        The anuria limit for urine output calculations.

    Example
    -------
    ```pycon
    >>> probe = UrineOutputProbe(column="urineoutput", anuria_limit=0.1)
    ... result_df = probe.probe(df=my_dataframe, patient=patient_df)
    ```
    """

    RESNAME = "urineoutput_stage"

    def __init__(
        self,
        column: str = "urineoutput",
        patient_weight_column: str = "weight",
        anuria_limit: float = 0.1,
        method: UrineOutputMethod = UrineOutputMethod.MEAN,
    ) -> None:
        super().__init__()

        self._column: str = column
        self._patient_weight_column: str = patient_weight_column

        self._anuria_limit: float = anuria_limit
        self._method: UrineOutputMethod = method

    @dataset_as_df(df=DatasetType.URINEOUTPUT, patient=DatasetType.DEMOGRAPHICS)
    @df_to_dataset(DatasetType.URINEOUTPUT)
    def probe(
        self,
        df: pd.DataFrame,
        patient: pd.DataFrame,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Perform urine output analysis on the provided DataFrame.

        This method calculates the KDIGO stage according to urine output based on the provided DataFrame and patient information DataFrame.
        It modifies the DataFrame by adding the urine output stage column with appropriate values based on the calculations.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame containing the urine output data. We expect the DataFrame to contain urine output values in ml, sampled hourly.
        patient : pd.DataFrame
            The DataFrame containing patient information. Should contain the patients weight in kg.

        Returns
        -------
        pd.DataFrame
            The modified DataFrame with the urine output stage column added.
        """
        if self._patient_weight_column not in patient:
            raise ValueError("Missing weight for stay")

        df = df.copy()

        weight: pd.Series = patient[self._patient_weight_column]
        # fmt: off
        df.loc[:, self.RESNAME] = np.nan  # set all urineoutput_stage values to NaN
        df.loc[df.rolling(6).min()[self._column] >= 0, self.RESNAME] = 0

        if self._method == UrineOutputMethod.STRICT:
            df.loc[(df.rolling(6).max()[self._column] / weight) < 0.5, self.RESNAME] = 1
            df.loc[(df.rolling(12).max()[self._column] / weight) < 0.5, self.RESNAME] = 2
            df.loc[(df.rolling(24).max()[self._column] / weight) < 0.3, self.RESNAME] = 3
            df.loc[(df.rolling(12).max()[self._column] / weight) < self._anuria_limit, self.RESNAME] = 3
        elif self._method == UrineOutputMethod.MEAN:
            df.loc[(df.rolling(6).mean()[self._column] / weight) < 0.5, self.RESNAME] = 1
            df.loc[(df.rolling(12).mean()[self._column] / weight) < 0.5, self.RESNAME] = 2
            df.loc[(df.rolling(24).mean()[self._column] / weight) < 0.3, self.RESNAME] = 3
            df.loc[(df.rolling(12).mean()[self._column] / weight) < self._anuria_limit, self.RESNAME] = 3
        else:
            raise ValueError(f"Invalid method: {self._method}")
        # fmt: on

        df.loc[pd.isna(df[self._column]), self.RESNAME] = np.nan

        return df


class CreatinineBaselineMethod(StrEnum):
    """
    Enumeration class representing different methods for creatinine baseline calculations.

    This class defines the available methods for calculating creatinine values.
    It is a subclass of the `StrEnum` class, which is a string-based enumeration.
    The available methods are `MIN` and `FIRST`.

    Attributes
    ----------
    ROLLING_MIN : str
        Minimum of a rolling window following the timepoint of observation is used as baseline.
    ROLLING_FIRST : str
        First value of a rolling window following the timepoint of observation is used as baseline.
    ROLLING_MEAN : str
        Mean of a rolling window following the timepoint of observation is used as baseline.
    FIXED_MIN : str
        Minimum of the first n days of observation is used as baseline.
    FIXED_MEAN : str
        Mean of the first n days of observation is used as baseline.
    OVERALL_FIRST : str
        First observed value is used as baseline.
    OVERALL_MEAN : str
        Mean of all observed values is used as baseline.
    OVERALL_MIN : str
        Minimum of all observed values is used as baseline.
    CONSTANT : str
        A constant value is used as baseline.
    CALCULATED : str
        A calculated value is used as baseline, based off of the Cockcroft-Gault-Formula using the Adjusted Body Weight.
    """

    ROLLING_MIN = auto()
    ROLLING_FIRST = auto()
    ROLLING_MEAN = auto()
    FIXED_MIN = auto()
    FIXED_MEAN = auto()
    OVERALL_FIRST = auto()
    OVERALL_MEAN = auto()
    OVERALL_MIN = auto()
    CONSTANT = auto()
    CALCULATED = auto()


class AbstractCreatinineProbe(Probe, metaclass=ABCMeta):
    """
    Abstract base class representing a creatinine probe.

    This class serves as an abstract base class for creatinine probes.
    It extends the `Probe` class and provides common functionality and attributes
    for creatinine probe implementations.

    Parameters
    ----------
    column : str, default: "creat"
        The name of the column containing creatinine values.
    baseline_constant_column : str, default: "baseline_constant"
        The name of the column containing constant baseline values.
    patient_weight_column : str, default: "weight"
        The name of the column containing the patient's weight.
    patient_age_column : str, default: "age"
        The name of the column containing the patient's age.
    patient_height_column : str, default: "height"
        The name of the column containing the patient's height.
    patient_gender_column : str, default: "gender"
        The name of the column containing the patient's gender.
    baseline_timeframe : str, default: "7d"
        The timeframe for calculating the baseline values.
    expected_clearance : float, default: 72
        The expected creatinine clearance rate.
    method : CreatinineBaselineMethod, default: CreatinineBaselineMethod.ROLLING_MIN
        The method for calculating the creatinine baseline values.

    Example
    -------
    ```pycon
    >>> class MyCreatinineProbe(AbstractCreCreatinineProbe):
    ...     def __init__(self, column="creatinine", baseline_timeframe="7d", method=CreatinineBaselineMethod.MIN):
    ...         super().__init__(column, baseline_timeframe, method)
    ...         # Additional initialization
    ...
    ...     def probe(self, df, **kwargs):
    ...         # Probe implementation specific to the derived class
    ```
    """

    def __init__(
        self,
        column: str = "creat",
        baseline_constant_column: str = "baseline_constant",
        patient_weight_column: str = "weight",
        patient_age_column: str = "age",
        patient_height_column: str = "height",
        patient_gender_column: str = "gender",
        baseline_timeframe: str = "7d",
        expected_clearance: float = 72,
        method: CreatinineBaselineMethod = CreatinineBaselineMethod.ROLLING_MIN,
    ) -> None:
        super().__init__()

        self._column: str = column
        self._baseline_constant_column: str = baseline_constant_column
        self._patient_weight_column: str = patient_weight_column
        self._patient_age_column: str = patient_age_column
        self._patient_height_column: str = patient_height_column
        self._patient_gender_column: str = patient_gender_column

        self._baseline_timeframe: str = baseline_timeframe
        self._expected_clearance: float = expected_clearance
        self._method: CreatinineBaselineMethod = method

    def creatinine_baseline(self, df: pd.DataFrame, patient: pd.DataFrame) -> pd.Series:
        """
        Calculate the creatinine baseline values.

        This method calculates the creatinine baseline values based on the configured
        parameters and the provided DataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame containing the creatinine data.

        Returns
        -------
        pd.Series
            The calculated creatinine baseline values.
        """
        if isinstance(df.index, PeriodIndex):
            df.index = df.index.to_timestamp()

        if self._method == CreatinineBaselineMethod.ROLLING_FIRST:
            return (
                df[df[self._column] > 0]
                .rolling(self._baseline_timeframe)
                .agg(lambda rows: rows.iloc[0])
                .resample("1h")
                .first()
                .ffill()[self._column]
            )

        if self._method == CreatinineBaselineMethod.ROLLING_MIN:
            return (
                df[df[self._column] > 0]
                .rolling(self._baseline_timeframe)
                .min()
                .resample("1h")
                .min()
                .ffill()[self._column]
            )
        if self._method == CreatinineBaselineMethod.ROLLING_MEAN:
            return (
                df[df[self._column] > 0]
                .rolling(self._baseline_timeframe)
                .mean()
                .resample("1h")
                .mean()
                .ffill()[self._column]
            )

        if self._method == CreatinineBaselineMethod.FIXED_MIN:
            values: pd.Series = (
                df[df[self._column] > 0]
                .rolling(self._baseline_timeframe)
                .min()
                .resample("1h")
                .min()
                .ffill()[self._column]
            )
            min_value: pd.DatetimeIndex = values[
                values.index <= (values.index[0] + pd.Timedelta(self._baseline_timeframe))
            ].min()  # calculate min value for first 7 days
            values[
                values.index > (values.index[0] + pd.Timedelta(self._baseline_timeframe))
            ] = min_value  # set all values after first 7 days to min value

            return values

        if self._method == CreatinineBaselineMethod.FIXED_MEAN:
            time_delta = pd.to_timedelta(self._baseline_timeframe)
            end_time = df.index[0] + time_delta
            value = df[df.index <= end_time][self._column].mean()
            values = self._to_df_length(df, value)
            return values

        if self._method == CreatinineBaselineMethod.OVERALL_FIRST:
            value = df[df[self._column] > 0].iloc[0][self._column]
            values = self._to_df_length(df, value)
            return values

        if self._method == CreatinineBaselineMethod.OVERALL_MIN:
            value = df[df[self._column] > 0][self._column].min()
            values = self._to_df_length(df, value)
            return values

        if self._method == CreatinineBaselineMethod.OVERALL_MEAN:
            value = df[df[self._column] > 0][self._column].mean()
            values = self._to_df_length(df, value)
            return values

        if self._method == CreatinineBaselineMethod.CONSTANT:
            if self._baseline_constant_column not in patient:
                raise ValueError(
                    "Baseline constant method requires baseline constant values. Please provide a pd.Series containing baseline values for creatinine."
                )

            return pd.Series(
                [patient[self._baseline_constant_column]] * len(df),
                index=df.index,
                name=self._column,
            )

        if self._method == CreatinineBaselineMethod.CALCULATED:
            columns = [
                self._patient_weight_column,
                self._patient_age_column,
                self._patient_height_column,
                self._patient_gender_column,
            ]
            for column in columns:
                if column not in patient:
                    raise ValueError(
                        f"Calculated baseline method requires patient {column}. Please provide a pd.Series containing patient {column}."
                    )

            weight = patient[self._patient_weight_column]
            height = patient[self._patient_height_column]
            gender = patient[self._patient_gender_column]
            age = patient[self._patient_age_column]

            ibw = (50.0 if gender == "M" else 45.5) + 2.3 * height / 2.54 - 60
            abw = ibw + 0.4 * (weight - ibw)

            # fmt: off
            return pd.Series(
                [
                    ((140 - age) * abw * (1 if gender == "M" else 0.85)) /
                    (70 * self._expected_clearance)
                ] * len(df),
                index=df.index,
                name=self._column,
            )
            # fmt: on

    def _to_df_length(self, df: pd.DataFrame, value: float) -> pd.Series:
        """
        Helper function to create a series, the same length as the data frame.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame to match the length of.
        value : float
            The value to fill the series with.

        Returns
        -------
        pd.Series
            The series with the same length as the DataFrame.
        """
        values = pd.Series(
            [value] * len(df),
            index=df.index,
            name=self._column,
        )
        return values


class AbsoluteCreatinineProbe(AbstractCreatinineProbe):
    """
    Probe class for absolute creatinine criterion, according to KDIGO criteria.

    This class represents a probe that calculates AKI stages according to absolute rises in creatinine, according to the KDIGO criteria.
    It extends the `AbstractCreCreatinineProbe` class.

    Attributes
    ----------
    RESNAME : str
        The name of the resulting stage column.

    Parameters
    ----------
    column : str, default: "creat"
        The name of the column containing creatinine values.
    baseline_constant_column : str, default: "baseline_constant"
        The name of the column containing constant baseline values.
    patient_weight_column : str, default: "weight"
        The name of the column containing the patient's weight.
    patient_age_column : str, default: "age"
        The name of the column containing the patient's age.
    patient_height_column : str, default: "height"
        The name of the column containing the patient's height.
    patient_gender_column : str, default: "gender"
        The name of the column containing the patient's gender.
    baseline_timeframe : str, default: "2d"
        The timeframe for calculating the baseline values.
    expected_clearance : float, default: 72
        The expected creatinine clearance rate.
    method : CreatinineBaselineMethod, default: CreatinineBaselineMethod.ROLLING_MIN
        The method for calculating the creatinine baseline values.

    Example
    -------
    ```pycon
    >>> probe = AbsoluteCreatinineProbe(column="creatinine", baseline_timeframe="7d", method=CreatinineMethod.MIN)
    ... df_result = probe.probe(df)
    ```
    """

    RESNAME = "abs_creatinine_stage"

    def __init__(
        self,
        column: str = "creat",
        baseline_constant_column: str = "baseline_constant",
        patient_weight_column: str = "weight",
        patient_age_column: str = "age",
        patient_height_column: str = "height",
        patient_gender_column: str = "gender",
        baseline_timeframe: str = "2d",
        expected_clearance: float = 72,
        method: CreatinineBaselineMethod = CreatinineBaselineMethod.ROLLING_MIN,
    ) -> None:
        super().__init__(
            column=column,
            baseline_constant_column=baseline_constant_column,
            patient_weight_column=patient_weight_column,
            patient_age_column=patient_age_column,
            patient_height_column=patient_height_column,
            patient_gender_column=patient_gender_column,
            baseline_timeframe=baseline_timeframe,
            expected_clearance=expected_clearance,
            method=method,
        )

    @dataset_as_df(df=DatasetType.CREATININE, patient=DatasetType.DEMOGRAPHICS)
    @df_to_dataset(DatasetType.CREATININE)
    def probe(
        self,
        df: pd.DataFrame,
        patient: pd.DataFrame,
        **kwargs: Any,
    ) -> pd.DataFrame:
        """
        Perform KDIGO stage calculation based on absolute creatinine elevations on the provided DataFrame.

        This method calculates the KDIGO stage based on the provided DataFrame
        and the configured baseline values. It calculates the stage according to KDIGO criteria.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame containing the creatinine data. It should have a column
            with the name specified in the `column` attribute of the probe.
        patient : pd.DataFrame
            The DataFrame containing patient information. Should contain the patients weight in kg and the age.

        Returns
        -------
        pd.DataFrame
            The modified DataFrame with the absolute creatinine stage column added.
        """
        df = df.copy()

        baseline_values: pd.Series = self.creatinine_baseline(df, patient)

        df.loc[:, self.RESNAME] = 0
        df.loc[approx_gte((df[self._column] - baseline_values), 0.3), self.RESNAME] = 1
        df.loc[approx_gte(df[self._column], 4), self.RESNAME] = 3

        df.loc[pd.isna(df[self._column]), self.RESNAME] = np.nan

        return df


class RelativeCreatinineProbe(AbstractCreatinineProbe):
    """
    Probe class for relative creatinine measurements.

    This class represents a probe calculates KDIGO stages based on relative creatinine elevations.

    Attributes
    ----------
    RESNAME : str
        The name of the resulting stage column.

    Parameters
    ----------
    column : str, default: "creat"
        The name of the column containing creatinine values.
    baseline_constant_column : str, default: "baseline_constant"
        The name of the column containing constant baseline values.
    patient_weight_column : str, default: "weight"
        The name of the column containing the patient's weight.
    patient_age_column : str, default: "age"
        The name of the column containing the patient's age.
    patient_height_column : str, default: "height"
        The name of the column containing the patient's height.
    patient_gender_column : str, default: "gender"
        The name of the column containing the patient's gender.
    baseline_timeframe : str, default: "7d"
        The timeframe for calculating the baseline values.
    expected_clearance : float, default: 72
        The expected creatinine clearance rate.
    method : CreatinineBaselineMethod, default: CreatinineBaselineMethod.ROLLING_MIN
        The method for calculating the creatinine baseline values.

    Example
    -------
    ```pycon
    >>> probe = RelativeCreatinineProbe(column="creatinine", baseline_timeframe="7d", method=CreatinineBaselineMethod.MIN)
    ... df_result = probe.probe(df)
    ```
    """

    RESNAME = "rel_creatinine_stage"

    @dataset_as_df(df=DatasetType.CREATININE, patient=DatasetType.DEMOGRAPHICS)
    @df_to_dataset(DatasetType.CREATININE)
    def probe(self, df: pd.DataFrame, patient: pd.DataFrame, **kwargs: Any) -> pd.DataFrame:
        """
        Perform calculation of relative creatinine elevations on the provided DataFrame.

        This method calculates the relative creatinine stage based on the provided DataFrame
        and the configured baseline values. It modifies the DataFrame by adding the relative
        creatinine stage column with appropriate values based on the calculations.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame containing the creatinine data. It should have a column
            with the name specified in the `column` attribute of the probe.
        patient : pd.DataFrame
            The DataFrame containing patient information. Should contain the patients weight in kg and the age.

        Returns
        -------
        pd.DataFrame
            The modified DataFrame with the relative creatinine stage column added.
        """
        df = df.copy()

        baseline_values: pd.Series = self.creatinine_baseline(df, patient)

        df.loc[:, self.RESNAME] = 0
        df.loc[approx_gte((df[self._column] / baseline_values), 1.5), self.RESNAME] = 1.0
        df.loc[approx_gte((df[self._column] / baseline_values), 2), self.RESNAME] = 2
        df.loc[approx_gte((df[self._column] / baseline_values), 3), self.RESNAME] = 3

        df.loc[pd.isna(df[self._column]), self.RESNAME] = np.nan

        return df


class RRTProbe(Probe):
    """
    Probe class for RRT.

    This class represents a probe that calculates RRT. It will return a KDIGO stage 3 if the patient is on RRT at any time during the ICU stay. It will return 0 otherwise.

    Attributes
    ----------
    RESNAME : str
        The name of the resulting stage column.

    Parameters
    ----------
    column : str, default: "rrt_status"
        The name of the column containing the RRT data.
    """

    RESNAME = "rrt_stage"

    def __init__(self, column: str = "rrt_status") -> None:
        super().__init__()

        self._column: str = column

    @dataset_as_df(df=DatasetType.RRT)
    @df_to_dataset(DatasetType.RRT)
    def probe(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Perform calculation of RRT on the provided DataFrame.

        Parameters
        ----------
        df : pd.DataFrame
            The DataFrame containing the RRT data. It should have a column
            with the name specified in the `column` attribute of the probe.

        Returns
        -------
        pd.DataFrame
            The modified DataFrame with the RRT stage column added.
        """
        df = df.copy()

        df.loc[:, self.RESNAME] = 0
        df.loc[df[self._column] == 1, self.RESNAME] = 3

        # transfer nans
        df.loc[pd.isna(df[self._column]), self.RESNAME] = np.nan

        return df
